/* anon-tar.c
 *
 * This file is part of the "anon-tar" package.
 *
 * The "anon-tar" package may be used under the terms and conditions of the
 * GNU General Public License version 2. A copy of these is included in the
 * file "COPYING" in the "doc" subdirectory.
 *
 * (C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>
 */

/* This program reads a GNUtar (or plain old us-tar) file on stdin and writes
 * it to stdout, replacing all user IDs and group IDs with 65534, all user
 * names with 'nobody' and all group names with 'nogroup'.
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>

static char *VERSION = "anon-tar 0.1  (C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>\n";

static char verbose = 0;

#define BLOCKSIZE 512

/* Posix magic string */
#define TMAGIC   "ustar\0"
#define TVERSION "00"

/* Entry types */
#define REGTYPE  '0'
#define AREGTYPE '\0'
#define LNKTYPE  '1'
#define SYMTYPE  '2'
#define CHRTYPE  '3'
#define BLKTYPE  '4'
#define DIRTYPE  '5'
#define FIFOTYPE '6'
#define CONTTYPE '7'

/* GNU expansion types */
#define GNUTYPE_DUMPDIR  'D'
#define GNUTYPE_LONGLINK 'K'
#define GNUTYPE_LONGNAME 'L'
#define GNUTYPE_MULTIVOL 'M'
#define GNUTYPE_NAMES    'N'
#define GNUTYPE_SPARSE   'S'
#define GNUTYPE_VOLHDR   'V'

/* Permission bits */
#define TSUID   04000
#define TSGID   02000
#define TSVTX   01000
#define TUREAD  00400
#define TUWRITE 00200
#define TUEXEC  00100
#define TGREAD  00040
#define TGWRITE 00020
#define TGEXEC  00010
#define TOREAD  00004
#define TOWRITE 00002
#define TOEXEC  00001

/* Old GNUtar magic string */
/* #define OLDGNU_MAGIC   "ustar  \0" */
#define OLDGNU_MAGIC   "ustar "
#define OLDGNU_VERSION " \0"

#define POSIX_HDR_name(block)     block
#define POSIX_HDR_mode(block)     (&block[100])
#define POSIX_HDR_uid(block)      (&block[108])
#define POSIX_HDR_gid(block)      (&block[116])
#define POSIX_HDR_size(block)     (&block[124])
#define POSIX_HDR_mtime(block)    (&block[136])
#define POSIX_HDR_chksum(block)   (&block[148])
#define POSIX_HDR_typeflag(block) (&block[156])
#define POSIX_HDR_linkname(block) (&block[157])
#define POSIX_HDR_magic(block)    (&block[257])
#define POSIX_HDR_version(block)  (&block[263])
#define POSIX_HDR_uname(block)    (&block[265])
#define POSIX_HDR_gname(block)    (&block[297])
#define POSIX_HDR_devmajor(block) (&block[329])
#define POSIX_HDR_devminor(block) (&block[337])
#define POSIX_HDR_prefix(block)   (&block[345])

#define GNU_HDR_atime(block)      (&block[345])
#define GNU_HDR_ctime(block)      (&block[357])
#define GNU_HDR_offset(block)     (&block[369])
#define GNU_HDR_longnames(block)  (&block[381])
#define GNU_HDR_sp(block)         (&block[386])
#define GNU_HDR_isextended(block) (&block[482])
#define GNU_HDR_realsize(block)   (&block[483])

#define EXTRA_HDR_atime(block)      block
#define EXTRA_HDR_ctime(block)      (&block[12])
#define EXTRA_HDR_offset(block)     (&block[24])
#define EXTRA_HDR_realsize(block)   (&block[36])
#define EXTRA_HDR_longnames(block)  (&block[48])
#define EXTRA_HDR_sp(block)         (&block[120])
#define EXTRA_HDR_isextended(block) (&block[504])

#define SPARSE_HDR_sp(block)         block
#define SPARSE_HDR_isextended(block) (&block[504])

static void usage(char *progname, int exitcode) {
char *USAGE = "Usage: ";
char *OPTIONS = " [-v] [-h | --help] [--version]\n";
char *OPT_V = "-v increases verbosity\n";

    write(STDERR_FILENO, USAGE, strlen(USAGE));
    write(STDERR_FILENO, progname, strlen(progname));
    write(STDERR_FILENO, OPTIONS, strlen(OPTIONS));
    write(STDERR_FILENO, OPT_V, strlen(OPT_V));

    exit(exitcode);
}

static unsigned int parseOct(unsigned char *str, int len) {
unsigned int value = 0, i;

    for (i = 0; i < len; i++) {
	if (str[i] < '0' || str[i] > '9') {
	    return value;
	}
	value *= 8;
	value += str[i] - '0';
    }
    return value;
}

static void checksum(unsigned char *block) {
unsigned int sum = 0, i;

    memset(POSIX_HDR_chksum(block), ' ', 8);

    for (i = 0; i < BLOCKSIZE; i++) {
	sum += block[i];
    }

    sum &= 0x3fff;

    POSIX_HDR_chksum(block)[6] = '\0';
    i = 6;
    do {
	i--;
	POSIX_HDR_chksum(block)[i] = '0' + (sum & 7);
	sum >>= 3;
    } while (i > 0);
}

static void logmsg(char *msg) {
    if (verbose) {
	write(STDERR_FILENO, msg, strlen(msg));
	write(STDERR_FILENO, "\n", 1);
    }
}

int main(int argc, char **argv) {
char *UNKN_OPTION = "Unknown option: ";
char *UNKN_HEADER = "Unknown header type! Aborting...\n";
char *INCOMPLETE_READ = "Incomplete read! Aborting...\n";
char *INCOMPLETE_WRITE = "Incomplete write! Aborting...\n";
char *ERR_WHILE_READING = "Error while reading: ";
char *ANON_UID = "0177776";
char *ANON_USER = "nobody";
char *ANON_GROUP = "nogroup";
int i, state = 0, size = 0, result, hdrIsPosix;
unsigned char block[BLOCKSIZE];
unsigned char emptyBlock[BLOCKSIZE];

    for (i = 1; i < argc; i++) {
	if (!strcmp(argv[i], "-v")) {
	    verbose = 1;
	} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
	    usage(argv[0], 0);
	} else if (!strcmp(argv[i], "--version")) {
	    write(STDOUT_FILENO, VERSION, strlen(VERSION));
	    exit(0);
	} else {
	    write(STDERR_FILENO, UNKN_OPTION, strlen(UNKN_OPTION));
	    write(STDERR_FILENO, argv[i], strlen(argv[i]));
	    write(STDERR_FILENO, "\n", 1);
	    usage(argv[0], 1);
	}
    }

    memset(emptyBlock, 0, BLOCKSIZE);
    while ((result = read(STDIN_FILENO, block, BLOCKSIZE)) > 0) {
        if (result < BLOCKSIZE) {
	    char *err = strerror(errno);
	    write(STDERR_FILENO, INCOMPLETE_READ, strlen(INCOMPLETE_READ));
	    write(STDERR_FILENO, err, strlen(err));
	    write(STDERR_FILENO, "\n", 1);
	    write(STDOUT_FILENO, block, result);
	    exit(1);
        }
        if (state == 0) {
	    if (memcmp(block, emptyBlock, BLOCKSIZE)) {
	        if (!memcmp(POSIX_HDR_magic(block), TMAGIC, 6) &&
			 !memcmp(POSIX_HDR_version(block), TVERSION, 2)) {
	            logmsg("Read posix header block");
		    hdrIsPosix = 1;
	            if (*POSIX_HDR_typeflag(block) != 0 &&
			    (*POSIX_HDR_typeflag(block) < '0'
			     || *POSIX_HDR_typeflag(block) > '9')) {
			/* Assume GNU extension header present */
		        state = 1;
	            }
	        } else if (!memcmp(POSIX_HDR_magic(block), OLDGNU_MAGIC, 6) &&
		           !memcmp(POSIX_HDR_version(block), OLDGNU_VERSION, 2)) {
	            logmsg("Read old gnu header block");
		    hdrIsPosix = 0;
	            if (*GNU_HDR_isextended(block) == '\001') {
		        state = 2;
	            }
	        } else {
	            write(STDERR_FILENO, UNKN_HEADER, strlen(UNKN_HEADER));
	            exit(1);
	        }
	        size = parseOct(POSIX_HDR_size(block), 12);
	        if (state == 0 && size > 0) { state = 3; }
		memcpy(POSIX_HDR_uid(block), ANON_UID, 8);
		memcpy(POSIX_HDR_gid(block), ANON_UID, 8);
		memset(POSIX_HDR_uname(block), 0, 32);
		memset(POSIX_HDR_gname(block), 0, 32);
		memcpy(POSIX_HDR_uname(block), ANON_USER, strlen(ANON_USER));
		memcpy(POSIX_HDR_gname(block), ANON_GROUP, strlen(ANON_GROUP));
		checksum(block);
	    }
        } else if (state == 1) {
	    logmsg("Read extra header block");
	    state = (*EXTRA_HDR_isextended(block) == '\001') ? 2 :
							       (size > 0 ? 3:0);
        } else if (state == 2) {
	    logmsg("Read sparse header block");
	    state = (*SPARSE_HDR_isextended(block) == '\001') ? 2 :
								(size >0 ? 3:0);
        } else if (state == 3) {
	    logmsg("Read data block");
	    size -= BLOCKSIZE;
	    if (size <= 0) {
	        state = 0;
	    }
        }
	if (write(STDOUT_FILENO, block, BLOCKSIZE) < BLOCKSIZE) {
	    char *err = strerror(errno);
	    write(STDERR_FILENO, INCOMPLETE_WRITE, strlen(INCOMPLETE_WRITE));
	    write(STDERR_FILENO, err, strlen(err));
	    write(STDERR_FILENO, "\n", 1);
	    exit(1);
	}
    }

    if (result < 0) {
	char *err = strerror(errno);
	write(STDERR_FILENO, ERR_WHILE_READING, strlen(ERR_WHILE_READING));
	write(STDERR_FILENO, err, strlen(err));
	write(STDERR_FILENO, "\n", 1);
	exit(1);
    }

    exit(0);
}

