# publish-freenet-mirror.sh: upload a given repository into freenet
# 
################################################################
# Copyright (C) 2001, 2002 Tom Lord
# Copyright (C) 2003 Peter Conrad
# (based on push-mirror.sh by Tom Lord)
# 
# See the file "COPYING" for further information about
# the copyright and warranty status of this work.
# 

set -e 
command_line="$*"

################################################################
# special options
# 
# Some options are special:
# 
#       --version | -V
#       --help | -h
# 
if test $# -ne 0 ; then

  for opt in "$@" ; do
    case $opt in

      --version|-V) exec larch --version
                    ;;


      --help|-h)
                printf "Upload a repository as a mirror into freenet\\n"
                printf "Usage: publish-freenet-mirror [options] from-archive to-archive\\n"
                printf "\\n"
                printf " -V --version                  print version info\\n"
                printf " -h --help                     display help\\n"
                printf "\\n"
                printf " --silent                      no output (except odd errors)\\n"
                printf " --quiet                       brief output\\n"
                printf " --report                      default output\\n"
                printf " --verbose                     maximal output\\n"
                printf " --debug                       debugging output\\n"
                printf "\\n"
                printf " --sskfile <file>              file containing SSK key pair\\n"
                printf " --all                         re-insert all files instead of only new ones\\n"
                printf " --new-edition                 increment edition counter before insert\\n"
                printf " --dbr <n>                     create a DBR to the (current+<n>)th edition\\n"
                printf " --dbr-offset <n>[h]           DBR offset to use (h: hours, otherwise seconds)\\n"
                printf " --dbr-increment <n>[h | d]    DBR increment to use (h: hours, d: days)\\n"
                printf "\\n"
		printf "Insert all categories, branches, versions and revisions in FROM-ARCHIVE\\n"
                printf "into freenet, optionally creating a new edition of TO-ARCHIVE.\\n"
                printf "Additionally, all files in the root directory of FROM-ARCHIVE are\\n"
                printf "inserted, too. This can be used for maintaining a freeseite at\\n"
                printf "the archive URI. Files in the root directory ending with *.tmpl will be\\n"
                printf "the archive URI. Files in the root directory of the form <base>.tmpl will be\\n"
                printf "copied to <base>, replacing \"$EDITION$\" with the current edition number\\n"
                printf "and \"$EDITION<n>$\" with the current edition number + <n>, for <n> in\\n"
                printf "{-3, -2, -1, +1, +2, +3}.\\n"
                printf "\\n"
                printf "In addition, the \"=meta-info\" directory of TO-ARCHIVE will contain a\\n"
                printf "file named \"mirror\" which contains the name of FROM-ARCHIVE.\\n"
                printf "\\n"
                printf "A manifest file will be inserted even if --new-edition is not\\n"
                printf "given. This may lead to collisions if not used carefully.\\n"
                printf "\\n"
                printf "The option --dbr will enable insertion of a redirect for the given period.\\n"
		printf "The options --dbr-offset and --dbr-increment specify the usual DBR parameters,\\n"
		printf "which will be stored in a file in the =meta-info directory in the archive.\\n"
		printf "If the file exists, values will be read from there instead of using those on\\n"
		printf "the command line.\\n"
                printf "\\n"
                printf "\\n"
                exit 0
                ;;

      *)
                ;;
    esac
  done
fi

################################################################
# Ordinary Options
# 
# 

quiet=--quiet
report=--report
verbose=
silent_opt=
quiet_opt=
report_opt=
verbose_opt=
debug_opt=
__restart=
ssk_secret=
all_files=
new_edition=
dbr=
dbr_offset=
dbr_increment=

while test $# -ne 0 ; do

  case "$1" in 

    ----restart) shift
                 __restart=----restart
                 ;;

    --silent)   shift
                quiet=
                report=
                verbose=
                silent_opt=--silent
                quiet_opt=
                report_opt=
                verbose_opt=
                ;;

    --quiet)    shift
                quiet=--quiet
                report=
                verbose=
                silent_opt=
                quiet_opt=--quiet
                report_opt=
                verbose_opt=
                ;;

    --report)   shift
                quiet=--quiet
                report=--report
                verbose=
                silent_opt=
                quiet_opt=
                report_opt=--report
                verbose_opt=
                ;;

    --verbose)  shift
                quiet=--quiet
                report=--report
                verbose=--verbose
                silent_opt=
                quiet_opt=
                report_opt=
                verbose_opt=--verbose
                ;;

    --debug)    shift
                larch heading "push-mirror: debugging output enabled\\n"
                set -x
                debug_opt=--debug
                ;;

    --sskfile)	shift
		ssk_secret="$1";
		shift
		;;

    --all)	shift
		all_files="yes";
		;;

    --new-edition)
		shift
		new_edition="yes";
		;;

    --dbr)	shift
		dbr=$1;
		shift
		;;

    --dbr-offset)
		shift
		case "$1" in
		    *h) dbr_offset=$((`echo $1 | sed 's=h=='` * 3600))
			;;
		    *)  dbr_offset="$1"
		esac
		shift
		;;

    --dbr-increment)
		shift
		case "$1" in
		    *h) dbr_increment=$((`echo $1 | sed 's=h=='` * 3600))
			;;
		    *d) dbr_increment=$((`echo $1 | sed 's=d=='` * 86400))
			;;
		    *)  dbr_increment="$1"
		esac
		shift
		;;

    --)                 shift
                        break
                        ;;
                        
    -*)                 printf "publish-freenet-mirror: unrecognized option (%s)\\n" "$1" 1>&2
                        printf "try --help\\n" 1>&2
                        exit 1
                        ;;

    *)                  break
                        ;;
  esac

done



################################################################
# Ordinary Arguments
# 

if test $# -lt 2 ; then 
  printf "usage: publish-freenet-mirror [options] from-archive to-archive\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

if test -z "$ssk_secret" ; then 
  printf "Must specify ssk key file!\\n" 1>&2
  printf "try --help\\n" 1>&2
  exit 1
fi

from_archive="$1"
shift

to_archive="$1"
shift

################################################################
# Sanity Check and Process Defaults
# 

larch valid-archive-name -e push-mirror "$from_archive"
larch valid-archive-name -e push-mirror "$to_archive"

################################################################
# Greetings
# 

if test "(" -z "$__restart" -a ! -z "$quiet" ")" -o ! -z "$report" ; then
  larch heading "publish-freenet-mirror\\n"
  printf "arguments: %s\\n" "$command_line" | fold -w 60 | larch body-indent
  larch heading --sub "publish-freenet-mirror start time: %s\\n" "`date`"
  if test ! -z "$report" ; then 
    larch heading --sub "from archive: %s\\n" "$from_archive"
    larch heading --sub "to archive: %s\\n" "$to_archive"
  fi
fi

from_archive_location=`larch whereis-archive "$from_archive"`
case "$from_archive_location" in
    /*)
	;;
    *)
	printf "publish-freenet-mirror currently can only publish local archives.\\n"
	exit 1;
	;;
esac

to_archive_location=`larch whereis-archive "$to_archive" | sed 's=//$=='`
case "$to_archive_location" in
    freenet:*)
	;;
    *)
	printf "publish-freenet-mirror: TO-ARCHIVE must be a freenet: URL\\n"
	exit 1;
	;;
esac

# How this works:
# - Every file to be included in the in-freenet mirror is handled as follows:
#   0. If the file is larger than 4095k exit with an error message.
#      (Later: insert splitfile)
#   1. If it's a .tar.gz file and not all files inside are marked as belonging
#      to nobody/nogroup, treat it with anon-tar
#   2. Generate its md5 sum
#   3. if the md5 sum is not in <archive>/=meta-info/freenet-keys, insert it
#      into freenet, recording the resulting CHK in
#      <archive>/=meta-info/freenet-keys
#   4. Generate a metadata entry of the form
#Document
#Name=<pathname-in-archive>
#Redirect.Target=<CHK>
#EndPart
#
# - Files to be treated like this are
#    + all files in the <archive> top-level directory
#    + all files in the <archive>/=meta-info directory except those starting
#      with "freenet-"
#    + in every category/branch/version/revision directory the files
#       * CONTINUATION
#       * log
#       * "$version--$lvl.patches.tar.gz"
#       * "$version--$lvl.src.tar.gz"
#       * "$version--$lvl.tar.gz"
#    + every category/[branch/[version/]]=README
#
# - If there is no file <archive>/=meta-info/mirror, create a metadata entry
#   with the name =meta-info/mirror and the CHK of =meta-info/name
#
# - If the <archive> top-level directory contains a file "index.html", create
#   a default Document entry with its CHK.
#
# - If <archive>/=meta-info/freenet-edition exists, read the edition number
#   from it, otherwise use edition number 0
#
# - Read <secret> SSK key from ssk key file
#
# - Insert the resulting metadata file as freenet:SSK@<secret>/<edition>
#   (Later: if metadata > 31k insert as CHK an create a redirect under SSK)
#
# - Increment edition and write it to <archive>/=meta-info/freenet-edition

################################################################
# Initialize temporary manifest file and other stuff
# 

here="`pwd`"
manifest="$here/,,publish-freenet-mirror.mf.$$"
filelist="$here/,,publish-freenet-mirror.fl.$$"
bail()
{
  cd "$here"
  rm -f "$manifest" "$filelist"*
  exit 1
}

trap "printf \"publish-freenet-mirror: interrupted -- cleaning up\\n\" 1>&2 ; bail" INT
rm -rf "$manifest"
cat > "$manifest" <<__EOH__
Version
Revision=1
__EOH__
rm -f "$filelist"
touch "$filelist"
touch "$from_archive_location/=meta-info/freenet-keys"

################################################################
# Handy Functions
#
 
# Treat $1 as described above. Don't insert it yet, but create a list of files
# for "fiq"
handle_file() {
    if test ! -f "$from_archive_location/$1"; then
	return 0
    fi

    if test ! -z "$verbose" ; then
        larch heading --sub --sub "handling: %s\\n" "$1"
    fi

    if test `find "$from_archive_location/$1" -size +4095k | wc -l` -gt 0; then
	printf "publish-freenet-mirror: file $from_archive_location/$1 too large - 4095k is current maximum\\n"
	bail
	exit 1
    fi

    case "$1" in
	*.tar.gz) if ! anonymize "$from_archive_location/$1"; then
		    return 1
		  fi
		  ;;
    esac

    md5=`md5sum "$from_archive_location/$1" | cut -c 1-32`
    if test `grep -c "^$md5" "$from_archive_location/=meta-info/freenet-keys"` = 0; then
        if test ! -z "$verbose" ; then
            larch heading --sub --sub --sub "listing: %s\\n" "$1"
        fi
	chk="CHK@"
	echo $md5 $chk >>"$from_archive_location/=meta-info/freenet-keys"
    else
	chk=`grep "^$md5" "$from_archive_location/=meta-info/freenet-keys" | cut -c 34-`
    fi
    echo $md5 "$chk $1" >>"$filelist"
}

# Check if $1 has already been anonymized and anonymize it if not
anonymize() {
    if test ! -z "$verbose" ; then
        larch heading --sub --sub --sub "anonymizing: %s\\n" "$1"
    fi
    if test `tar tvfz "$1" | grep -vE '^.......... nobody/nogroup ' | wc -l` -gt 0; then
	tmp=`dirname "$1"`/,,`basename "$1"`.tmp.$$
	if gzip -d <"$1" | anon-tar | gzip -9 >"$tmp"; then
	    mv "$tmp" "$1"
	else
	    rm -f "$tmp"
	    printf "publish-freenet-mirror: anon-tar exited with an error\\n"
	    bail
	fi
    fi
}

################################################################
# Upload each file from the list described above
# 

edition=-1
if test -f "$from_archive_location/=meta-info/freenet-edition"; then
    edition=`cat "$from_archive_location/=meta-info/freenet-edition"`
fi
if [ "$new_edition" = "yes" ]; then
    edition=$(($edition + 1))
fi

# all files in <archive> top level dir
ls -a "$from_archive_location" | \
    grep -vE '^(\.|\.\.|)$' | \
    while read item; do
	case "$item" in
	    *.tmpl)
		converted=`basename "$item" .tmpl`
		sed < "$from_archive_location/$item" \
		    > "$from_archive_location/$converted" \
		    "s=\\\$EDITION-3\\\$=$(($edition - 3))=g;s=\\\$EDITION-2\\\$=$(($edition - 2))=g;s=\\\$EDITION-1\\\$=$(($edition - 1))=g;s=\\\$EDITION\\\$=$edition=g;s=\\\$EDITION+1\\\$=$(($edition + 1))=g;s=\\\$EDITION+2\\\$=$(($edition + 2))=g;s=\\\$EDITION+3\\\$=$(($edition + 3))=g"
		handle_file "$converted" || bail
		;;
	    *)  if [ ! -r "$item.tmpl" ]; then
		    handle_file "$item" || bail
		fi
	esac
    done

(cd "$from_archive_location"; \
 find "=meta-info" -type f ! -name "freenet-*") |
    while read item; do
	handle_file "$item" || bail
    done

larch categories -A "$from_archive" | \
  while read category; do
    if test ! -z "$quiet" ; then
      larch heading --sub "handling: %s\\n" "$category"
    fi
    handle_file "$category/=README" || bail
    larch branches -A "$from_archive" "$category" | \
    while read branch; do
      if test ! -z "$report" ; then
        larch heading --sub "handling: %s\\n" "$branch"
      fi
      handle_file "$category/$branch/=README" || bail
      larch versions -A "$from_archive" "$branch" | \
      while read version; do
        if test ! -z "$verbose" ; then
          larch heading --sub "handling: %s\\n" "$version"
        fi
	handle_file "$category/$branch/$version/=README" || bail
        larch revisions -A "$from_archive" "$version" | \
        while read revision; do
          if test ! -z "$verbose" ; then
            larch heading --sub "handling: %s\\n" "$revision"
          fi
	  handle_file "$category/$branch/$version/$revision/CONTINUATION" || bail
	  handle_file "$category/$branch/$version/$revision/log" || bail
	  handle_file "$category/$branch/$version/$revision/$version--$revision.patches.tar.gz" || bail
	  handle_file "$category/$branch/$version/$revision/$version--$revision.src.tar.gz" || bail
	  handle_file "$category/$branch/$version/$revision/$version--$revision.tar.gz" || bail
        done || bail
      done || bail
    done || bail
  done || bail

if test "$all_files" = "yes"; then
    if test ! -z "$quiet" ; then
        larch heading "Uploading files... this may take a while\\n"
    fi
    cp "$filelist" "$filelist.1"
    fiq -d "$from_archive_location" -t 8 "$filelist.1" "$filelist.out" || bail;
elif grep "CHK@ " "$filelist" >"$filelist.1"; then
    if test ! -z "$quiet" ; then
        larch heading "Uploading files... this may take a while\\n"
    fi
    fiq -d "$from_archive_location" -t 8 "$filelist.1" "$filelist.out" || bail;
elif test ! -z "$quiet" ; then
    larch heading "No files to upload.\\n"
fi

while read md5 chk file; do
    if [ "$chk" = "CHK@" ]; then
	chk=`grep "^$md5 " "$filelist.out" | cut -c 34-`
    fi
    if [ `grep -c "^$md5 CHK@\$" "$from_archive_location/=meta-info/freenet-keys"` -ge 1 ]; then
	mv "$from_archive_location/=meta-info/freenet-keys" "$from_archive_location/=meta-info/freenet-keys.old"
	sed <"$from_archive_location/=meta-info/freenet-keys.old" >"$from_archive_location/=meta-info/freenet-keys" "s=^$md5 CHK@=$md5 $chk="
    fi
    mimetype=
    case "$file" in
	*.txt) mimetype=text/plain
	    ;;
	*.html) mimetype=text/html
	    ;;
	*.gif) mimetype=image/gif
	    ;;
	*.png) mimetype=image/png
	    ;;
	*.jpg|*.jpeg) mimetype=image/jpeg
	    ;;
	*.pdf) mimetype=application/pdf
	    ;;
    esac
    if [ "$mimetype" = "" ]; then
	cat <<__EOP__ >>"$manifest"
EndPart
Document
Name=$file
Redirect.Target=$chk
__EOP__
    else
	cat <<__EOP__ >>"$manifest"
EndPart
Document
Info.Format=$mimetype
Name=$file
Redirect.Target=$chk
__EOP__
    fi
done < "$filelist"

if test -f "$from_archive_location/index.html"; then
    md5=`md5sum "$from_archive_location/index.html" | cut -c 1-32`
    chk=`grep "^$md5" "$from_archive_location/=meta-info/freenet-keys" | cut -c 34-`
    if test ! -z "$chk"; then
	cat >>$manifest <<__EOP__
EndPart
Document
Info.Format=text/html
Redirect.Target=$chk
__EOP__
    fi
fi

if test -f "$from_archive_location/=meta-info/name"; then
    md5=`md5sum "$from_archive_location/=meta-info/name" | cut -c 1-32`
    chk=`grep "^$md5" "$from_archive_location/=meta-info/freenet-keys" | cut -c 34-`
    if test ! -z "$chk"; then
	cat >>$manifest <<__EOP__
EndPart
Document
Name==meta-info/mirror
Redirect.Target=$chk
__EOP__
    fi
fi
echo "End" >>$manifest

name=`echo "$to_archive_location" | sed 's=[^/]*/=='`
if test ! -z "$quiet" ; then
    larch heading --sub "Uploading meta info for freenet:SSK@<secret>/$name/$edition\\n"
fi
if [ `cat "$manifest" | wc -c` -gt 32000 ]; then
    # Need to insert as CHK, then redirect...
    if test ! -z "$quiet" ; then
	larch heading --sub --sub "Meta info > 32k, inserting as CHK first\\n"
    fi
    echo `md5sum "$manifest"` CHK@ "$manifest" >"$filelist"
    fiq "$filelist" "$filelist.out" || bail;
    read md5 chk < "$filelist.out"
    cat >$manifest <<__EOM__
Version
Revision=1
EndPart
Document
Redirect.Target=$chk
End
__EOM__
fi
ssktool --insert /dev/null --metadata "$manifest" \
	--sskpath "$name/$edition" --sskfile "$ssk_secret" 

if test "$new_edition" = "yes"; then
    echo $edition >"$from_archive_location/=meta-info/freenet-edition"
fi

cd "$here"
if [ "$dbr" != "" ]; then
    dbrinfo="$from_archive_location/=meta-info/freenet-dbrinfo"
    if [ -r "$dbrinfo" ]; then
	dbr_offset=`grep "^DBR_Offset" "$dbrinfo" | cut -d\  -f 2`
	dbr_increment=`grep "^DBR_Increment" "$dbrinfo" | cut -d\  -f 2`
    else
	if [ "$dbr_offset" = "" ]; then
	    dbr_offset=0
	fi
	if [ "$dbr_increment" = "" ]; then
	    dbr_increment=86400
	fi
	echo DBR_Offset: $dbr_offset >"$dbrinfo"
	echo DBR_Increment: $dbr_increment >>"$dbrinfo"
    fi
    if test ! -z "$quiet" ; then
	larch heading --sub "Uploading DBR redirect freenet:SSK@<secret>/$offset-$name\\n"
    fi
    ssktool --gendbr "$to_archive_location/$edition" --dbr-period $dbr \
	    --dbr-offset $dbr_offset --dbr-increment $dbr_increment \
	    --sskfile "$ssk_secret" --sskpath "$name"
fi
rm -rf "$manifest" "$filelist"*

if test ! -z "$quiet" ; then 
  larch heading --sub "publish-freenet-mirror finish time: %s\\n" "`date`"
fi

# tag: publish-freenet-mirror.sh 3515803b24d52f086929d9d5eb3e65f7
#
