/* This file is part of the "fiq" package.
 *
 * This package may be used under the terms and conditions of the GNU General
 * Public License version 2. A copy of the license can be found in the "doc"
 * subdirectory of the package distribution.
 *
 * (C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "worker.h"
#include "fiq.h"

void die(char *msg) {
    fprintf(stderr, "%s\nError was %s\n", msg, strerror(errno));
    exit(-1);
}

static char *me;
static void usage(char *msg) {
    if (msg != NULL) {
	fprintf(stderr, "%s\n", msg);
    }
    fprintf(stderr, "Usage: %s [options] <inputfile> <outputfile>\n", me);
    fprintf(stderr, "Available options are:\n");
    fprintf(stderr, " -d <dir>\tchange into <directory> before doing anything else\n");
    fprintf(stderr, " -t <num>\tspawn <num> insert threads (default: 1)\n");
    fprintf(stderr, " -h\tdisplay this help\n");
    fprintf(stderr, " --help\tdisplay this help\n");
    fprintf(stderr, " --version\tdisplay version number\n");

    exit(msg == NULL ? 0 : -1);
}

int main(int argc, char **argv) {
int i = 1, j, *pids;
int threads = 1, success = 1;
struct stat stat_buf;

    me = argv[0];
    while (i < argc - 2) {
	if (!strcmp(argv[i], "-d")) {
	    if (chdir(argv[++i])) {
		die("Couldn't change to directory!");
	    }
	} else if (!strcmp(argv[i], "-t")) {
	    threads = atoi(argv[++i]);
	    if (threads < 1) {
		fprintf(stderr, "You must specify at least 1 insert thread!\n");
		exit(-1);
	    }
	} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
	    usage(NULL);
	} else if (!strcmp(argv[i], "--version")) {
	    fprintf(stderr, "%s version %s\n", me, VERSION);
	    fprintf(stderr, "(C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>\n");
	    fprintf(stderr, "This software may be used and / or distributed under the terms and conditions\n");
	    fprintf(stderr, "of the GNU General Public License version 2.\n");
	    exit(0);
	} else {
	    usage("Unknown option!");
	}
	i++;
    }

    if (i != argc - 2) {
	usage("Please specify <inputfile> and <outputfile>!");
    }

    if (stat(argv[i], &stat_buf)) {
	die("Can't stat input file!");
    }

    if (threads > 1) {
        pids = malloc(sizeof(int *) * threads);
        if (!pids) {
	    die("Malloc failed!");
        }

        for (j = 0; j < threads; j++) {
	    pids[j] = fork();
	    if (pids[j] < 0) { // error
	        fprintf(stderr, "fork() failed - %s\n", strerror(errno));
	        fprintf(stderr, "Continuing anyway...\n");
	    } else if (pids[j] == 0) { // child
	        worker(argv[i], argv[i+1]);
	        exit(0);
	    }
	}

        for (j = 0; j < threads; j++) {
	    if (pids[j] > 0) {
	        int rc, status;
	        do {
		    rc = waitpid(pids[j], &status, 0);
	        } while (rc != pids[j] && (rc != -1 || errno == EINTR));
	        if (rc < 0) {
		    die("Error while waiting for child!");
	        }
	        if (!WIFEXITED(status) || WEXITSTATUS(status)) {
		    success = 0;
	        }
	    }
        }

        free(pids);
    } else {
	worker(argv[i], argv[i+1]);
    }

    fprintf(stderr, "Done.\n");

    exit(success ? 0 : -1);
}

