/* This file is part of the "fiq" package.
 *
 * This package may be used under the terms and conditions of the GNU General
 * Public License version 2. A copy of the license can be found in the "doc"
 * subdirectory of the package distribution.
 *
 * (C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>
 */

#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h> 
#include <sys/stat.h>
#include <sys/wait.h>

#include <ezFCPlib/ezFCPlib.h>

#include "worker.h"
#include "fiq.h"

static char *putfile(HFCP *handle, char *chk, char *file) {
    if (fcpPutKeyFromFile(handle, chk, file, NULL)) {
	die("Insert failed!");
    }

    return handle->created_uri;
}

void worker(char *infile, char *outfile) {
int in, out, readBufLen = 512, readBufUsed, toRead, startOfLine, repeat;
struct stat stat_buf;
char *readBuf, *md5, *chk, *file, *newChk;
char *fcpHost, *tmp;
int  htl = 15, port = 8481;
HFCP *handle;

    fcpHost = getenv("FCP_HOST");
    if (fcpHost == NULL) { fcpHost = "localhost"; }
    tmp = getenv("FCP_PORT");
    if (tmp) {
	port = atoi(tmp);
    }
    tmp = getenv("INSERT_HTL");
    if (tmp) {
	htl = atoi(tmp);
    }
    if (fcpStartup(fcpHost, port, htl, 1, 1)) {
	die("fcpStartup failed!");
    }
    if (!(handle = fcpCreateHandle())) {
	die("fcpCreateHandle failed!");
    }
    /* Set RemoveLocalKey in FCP put message, so repeated inserts improve
     * availability */
    fcpSetRmLocal(handle, 1);
    /* Hack: try avoiding splits for now */
    fcpSetSplitSize(4000000);

    readBuf = malloc(readBufLen);
    if (!readBuf) {
	die("Can't malloc readBuf!");
    }

    in = open(infile, O_RDWR);
    if (in < 0) {
	die("Can't open input file!");
    }
    out = open(outfile, O_CREAT | O_WRONLY | O_APPEND, 0666);
    if (out < 0) {
	die("Can't open output file!");
    }

    while(1) { /*the loop will close the files and return when infile is empty*/
	if (flock(in, LOCK_EX)) {
	    die("Couldn't lock input file!");
	}
	if (fstat(in, &stat_buf)) {
	    die("Couldn't fstat input file!");
	}
	if (stat_buf.st_size == 0) {
	    flock(in, LOCK_UN);
	    close(in);
	    close(out);
	    fcpDestroyHandle(handle);
	    return;
	}
	do {
	    toRead = readBufLen - 1;
	    if (stat_buf.st_size < toRead) { toRead = stat_buf.st_size; }
	    if (lseek(in, stat_buf.st_size - toRead, SEEK_SET)
		    != stat_buf.st_size - toRead) {
	        die("Couldn't lseek input file!");
	    }
	    if (read(in, readBuf, toRead) != toRead) {
	        die("Short read!");
	    }
	    readBufUsed = toRead;
	    if (readBuf[readBufUsed-1] == '\n') { readBufUsed--; }
	    if (readBufUsed > 0
		&& readBuf[readBufUsed-1] == '\r') { readBufUsed--; }
	    startOfLine = (readBufUsed == 0 ? 0 : readBufUsed-1);
	    while (startOfLine > 0 && readBuf[startOfLine - 1] != '\n'
		   && readBuf[startOfLine - 1] != '\r') {
		startOfLine--;
	    }
	    repeat = (toRead == readBufLen - 1 && startOfLine == 0);
	    if (repeat) {
		readBufLen *= 2;
		readBuf = realloc(readBuf, readBufLen);
		if (!readBuf) {
		    die("realloc failed!");
		}
	    }
	} while (repeat);
	if (ftruncate(in, stat_buf.st_size - (toRead - startOfLine))) {
	    die("Couldn't truncate input file!");
	}
	if (flock(in, LOCK_UN)) {
	    die("Couldn't unlock input file!");
	}

	readBuf[readBufUsed] = 0;
	md5 = &readBuf[startOfLine];
	chk = md5;
	while (*chk != ' ' && *chk != 0) { chk++; }
	if (*chk == 0) {
	    fprintf(stderr, "Bad line: %s\n", md5);
	    close(in);
	    close(out);
	    exit(-1);
	}
	*chk = 0;
	chk++;
	file = chk;
	while (*file != ' ' && *file != 0) { file++; }
	if (*file == 0) {
	    fprintf(stderr, "Bad line: %s\n", md5);
	    close(in);
	    close(out);
	    exit(-1);
	}
	*file = 0;
	file++;

	newChk = putfile(handle, chk, file);

	if (flock(out, LOCK_EX)) {
	    die("Couldn't lock output file!");
	}
	if (fstat(out, &stat_buf)) {
	    die("Couldn't fstat output file!");
	}
	/*
	if (lseek(out, stat_buf.st_size, SEEK_SET) != stat_buf.st_size) {
	    die("Couldn't lseek output file!");
	}
	*/
	if (write(out, md5, strlen(md5)) != strlen(md5)) {
	    die("write failed!");
	}
	if (write(out, " ", 1) != 1) {
	    die("write failed!");
	}
	if (write(out, newChk, strlen(newChk)) != strlen(newChk)) {
	    die("write failed!");
	}
	if (write(out, "\n", 1) != 1) {
	    die("write failed!");
	}
	if (flock(out, LOCK_UN)) {
	    die("Couldn't unlock output file!");
	}
    }
}

/** Write log messages to stderr */
static void fn_debug(char *msg) {
    write(STDERR_FILENO, msg, strlen(msg));
    fsync(STDERR_FILENO);
}

/** Callback function used by ezFCPlib
 */
int fcpLogCallback(int level, char *buf) {
char *severity = "";
 
    if (level == FCP_LOG_CRITICAL) {
        severity = "CRITICAL";
    } else if (level == FCP_LOG_NORMAL) {
        severity = "NORMAL";
    } else if (level == FCP_LOG_VERBOSE) {
        severity = "VERBOSE";
    } else if (level == FCP_LOG_DEBUG) {
        severity = "DEBUG";
    }

    if (level == FCP_LOG_CRITICAL || level == FCP_LOG_NORMAL) { 
        fn_debug("ezFCPlib [");
        fn_debug(severity);
        fn_debug("]: ");
        fn_debug(buf);
        fn_debug("\n");
    }
 
    return 0;
}

