/* This file is part of the "fiq" package.
 *
 * This package may be used under the terms and conditions of the GNU General
 * Public License version 2. A copy of the license can be found in the "doc"
 * subdirectory of the package distribution.
 *
 * (C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <ezFCPlib/ezFCPlib.h>

#define VERSION	"0.1"

static const int CMD_GENKEYS  = 1;
static const int CMD_REDIRECT = 2;
static const int CMD_GENDBR   = 3;
static const int CMD_INSERT   = 4;

void die(char *msg) {
    fprintf(stderr, "%s\nError was %s\n", msg, strerror(errno));
    exit(-1);
}

static char *me;
static void usage(char *msg) {
    if (msg != NULL) {
	fprintf(stderr, "%s\n", msg);
    }
    fprintf(stderr, "Usage: %s [options] [command]\n", me);
    fprintf(stderr, "\nAvailable options are:\n");
    fprintf(stderr, " --sskfile <filename>   file with ssk keypair\n");
    fprintf(stderr, " --sskpath <path>       path fragment to append to SSK key in insert URI\n");
    fprintf(stderr, " --dbr-offset <time>    DBR offset to use (<n>[d|h] for days or hours)\n");
    fprintf(stderr, " --dbr-increment <time> DBR increment to use (<n>[d|h] for days or hours)\n");
    fprintf(stderr, " --dbr-period <x>       the DBR period offset for --gendbr\n");
    fprintf(stderr, " --content-type <str>   the content type for inserts\n");
    fprintf(stderr, " --metadata <file>      the meta data for inserts\n");
    fprintf(stderr, " --key <key>            the key for inserts\n");
    fprintf(stderr, "\nAvailable commands are\n");
    fprintf(stderr, " --genkeys             generate keypair and write to sskfile\n");
    fprintf(stderr, " --gendbr <target-URI> generate a DBR redirecting to <target-URI>\n");
    fprintf(stderr, " --redirect <target>   generate a simple redirect to <target>\n");
    fprintf(stderr, " --insert <file>       insert the given file\n");
    fprintf(stderr, " --version             display version number\n");
    fprintf(stderr, " --help | -h           show this help\n");
    fprintf(stderr, "\nThe commands --genkeys, --gendbr and --redirect require the --sskfile option.\n");
    fprintf(stderr, "The commands --gendbr and --redirect require the --sskpath option.\n");
    fprintf(stderr, "--redirect will insert metadata \"Redirect.Target=<target>\" under the URI\n");
    fprintf(stderr, "freenet:SSK@<publickey>PAgM/<sskpath>. --gendbr will insert  a \"base DBR\"\n");
    fprintf(stderr, "as freenet:SSK@<publickey>PAgM/<sskpath> with the given offset and increment\n");
    fprintf(stderr, "AND a redirect to <target> as freenet:SSK@<publickey>PAgM/<timestamp>-<sskpath>\n");
    fprintf(stderr, "where <timestamp> is the currently active timestamp plus <x> times <increment>.\n");
    fprintf(stderr, "--insert allows the specification of an alternative key instead of --sskfile\n");
    fprintf(stderr, "and --sskpath. Also, an optional meta info file can be specifiead, or\n");
    fprintf(stderr, "alternatively the content type.\n");

    exit(msg == NULL ? 0 : -1);
}

/** Generate a key pair and write it to sskfile */
static void genKey(HFCP *hfcp, char *sskfile) {
int  fd;
struct stat sb;
char pubkey[256], privkey[256];
char *PUBKEY  = "Public key:  ";
char *PRIVKEY = "Private key: ";

    if (!sskfile) {
	die("--genkeys requires --sskfile!");
    }
    if (!stat(sskfile, &sb)) {
	die("sskfile exists! Will not overwrite!");
    }
    if (fcpMakeSvkKeypair(hfcp, pubkey, privkey)) {
	die("fcpMakeSvkKeypair failed!");
    }
    if ((fd=open(sskfile, O_CREAT | O_EXCL | O_WRONLY, S_IRUSR | S_IWUSR)) < 0){
        fcpDestroyHandle(hfcp);
	die("Couldn't open sskfile!");
    }
    if (write(fd, PUBKEY, strlen(PUBKEY)) != strlen(PUBKEY)
	    || write(fd, pubkey, strlen(pubkey)) != strlen(pubkey)
	    || write(fd, "\n", 1) != 1
	    || write(fd, PRIVKEY, strlen(PRIVKEY)) != strlen(PRIVKEY)
	    || write(fd, privkey, strlen(privkey)) != strlen(privkey)
	    || write(fd, "\n", 1) != 1) {
	die("Incomplete write!");
    }
    close(fd);
}

/** Read keys from the given file. pubkey and privkey should be free()d later.*/
static void readKeys(char *sskfile, char **pubkey, char **privkey) {
char buffer[10240], *newline;
int  fd, count;

    if (!sskfile) {
	die("sskfile parameter missing!");
    }
    if (!pubkey || !privkey) {
	die("Illegal argument: NULL");
    }

    if ((fd=open(sskfile, O_RDONLY)) < 0) {
	die("Can't open sskfile!");
    }
    if ((count = read(fd, buffer, sizeof(buffer) - 1)) < 0) {
	die("read() from sskfile failed?!");
    }
    close(fd);
    buffer[count] = 0;
    if (strncmp("Public key:  ", buffer, 13)) {
	die("sskfile has wrong format!");
    }
    newline = strchr(buffer, '\n');
    if (!newline || strncmp("\nPrivate key: ", newline, 14)) {
	die("sskfile has wrong format!");
    }
    *newline = 0;
    *pubkey  = strdup(buffer + 13);
    if (!*pubkey) {
	die("strdup failed!");
    }
    *privkey = strdup(newline + 14);
    if (!*privkey) {
	die("strdup failed!");
    }
    for (count = 0; (*privkey)[count] && (*privkey)[count] != '\r'
		    &&  (*privkey)[count] != '\n'; count++);
    (*privkey)[count] = 0;
/*
    write(STDERR_FILENO, "Pubkey: ", 8);
    write(STDERR_FILENO, *pubkey, strlen(*pubkey));
    write(STDERR_FILENO, ",\nPrivkey: ", 11);
    write(STDERR_FILENO, *privkey, strlen(*privkey));
    write(STDERR_FILENO, ".\n", 2);
*/
}

/** Create a redirect from freenet:SSK@pubkey/sskpath to target */
static void genRedirect(HFCP *hfcp, char *sskfile, char *sskpath, char *target){
char *pubkey, *privkey;
char *meta, *uri;
char *META_1="Version\nRevision=1\nEndPart\nDocument\nRedirect.Target=";
char *META_2="\nEnd\n";
char *URI_1="freenet:SSK@";
char *URI_2="/";

    readKeys(sskfile, &pubkey, &privkey);

    if (!sskpath) {
	die("sskpath missing!");
    }
    if (!target) {
	die("<target> missing!");
    }

    uri = malloc(strlen(URI_1) + strlen(privkey)
		 + strlen(URI_2) + strlen(sskpath) + 1);
    if (!uri) {
	die("Can't malloc!");
    }
    meta = malloc(strlen(META_1) + strlen(target) + strlen(META_2) + 1);
    if (!meta) {
	die("Can't malloc!");
    }

    strcpy(meta, META_1);
    strcat(meta, target);
    strcat(meta, META_2);

    strcpy(uri, URI_1);
    strcat(uri, privkey);
    strcat(uri, URI_2);
    strcat(uri, sskpath[0] == '/' ? sskpath + 1 : sskpath);

    if (fcpPutKeyFromMem(hfcp, uri, uri, meta, 0)) {
        fcpDestroyHandle(hfcp);
	die("Insert failed!");
    }

    free(meta);
    free(uri);
    free(pubkey);
    free(privkey);
}

/** Generate a base DBR for the given offset, increment and path
 *  AND a date-specific redirect for the given period to the given target
 */
static void genDBR(HFCP *hfcp, char *sskfile, char *sskpath, char *target,
		   int dbr_increment, int dbr_offset, int dbr_period) {
char *pubkey, *privkey;
int rc;

    readKeys(sskfile, &pubkey, &privkey);

    if (!sskpath) {
	die("sskpath missing!");
    }
    if (!target) {
	die("<target> missing!");
    }
    while (*sskpath == '/') { sskpath++; }

    if (dbr_increment < 0) { dbr_increment = -dbr_increment; }
    if (dbr_increment == 0) { dbr_increment = 86400; }
    while (dbr_offset < 0) { dbr_offset += dbr_increment; }

    if (!(rc = fork())) {
	/* child - insert redirect */
	time_t now = time(NULL), then;
	char *dpath = malloc(strlen(sskpath) + 20);
	/*write(STDERR_FILENO, "child forked\n", 13);*/
	if (!dpath) {
	    die("malloc failed!");
	}
	/*write(STDERR_FILENO, "malloc successful\n", 18);*/
	then = now - (now % dbr_increment);
	then += dbr_offset;
	if (then > now) { then -= dbr_increment; }
	then += dbr_increment * dbr_period;
	/*write(STDERR_FILENO, "time calculated\n", 16);*/
	sprintf(dpath, "%lx-%s", then, sskpath);
	/*write(STDERR_FILENO, "calling redirect\n", 17);*/
	genRedirect(hfcp, sskfile, dpath, target);
	free(dpath);
	exit(0);
    } else if (rc > 0) {
	/* parent - insert base and wait for child */
	char *meta, *uri;
	char *META_1="Version\nRevision=1\nEndPart\nDocument\nDateRedirect.Target=freenet:SSK@";
	char *META_2="PAgM/";
	char *META_3="\nDateRedirect.Offset=";
	char *META_4="\nDateRedirect.Increment=";
	char *META_5="\nEnd\n";
	char *URI_1="freenet:SSK@";
	char *URI_2="/";
	int  ignore, pid;

	meta = malloc(strlen(META_1) + strlen(pubkey) + strlen(META_2)
		      + strlen(sskpath) + strlen(META_3) + 20
		      + strlen(META_4) + 20 + strlen(META_5) + 1);
	if (!meta) {
	    die("malloc failed!");
	}
	uri = malloc(strlen(URI_1) + strlen(privkey)
		     + strlen(URI_2) + strlen(sskpath) + 1);
	if (!uri) {
	    die("malloc failed!");
	}

	strcpy(meta, META_1);
	strcat(meta, pubkey);
	strcat(meta, META_2);
	strcat(meta, sskpath);
	if (dbr_offset) {
	    strcat(meta, META_3);
	    sprintf(meta + strlen(meta), "%x", dbr_offset);
	}
	if (dbr_increment && dbr_increment != 86400) {
	    strcat(meta, META_4);
	    sprintf(meta + strlen(meta), "%x", dbr_increment);
	}
	strcat(meta, META_5);
	
        strcpy(uri, URI_1);
        strcat(uri, privkey);
        strcat(uri, URI_2);
        strcat(uri, sskpath);

        if (fcpPutKeyFromMem(hfcp, uri, uri, meta, 0)) {
            fcpDestroyHandle(hfcp);
	    die("Insert failed!");
        }

	free(uri);
	free(meta);

	do {
	    pid = waitpid(rc, &ignore, 0);
	} while (pid > 0 && pid != rc);
	if (pid < 0) {
	    die("waitpid failed!");
	}
    } else {
	die("fork failed!");
    }

    free(pubkey);
    free(privkey);
}

/** Insert the given file either under the given key or under ssk.pub/sskpath,
 *  optionally using the supplied meta data or content type.
 */
static void insertFile(HFCP *hfcp, char *sskfile, char *sskpath, char *key,
		       char *file, char *ct, char *meta) {
int metalen = 0;
char *metadata = NULL;
struct stat sb; 
char *META_1 = "Version\nRevision=1\nEndPart\nDocument\nInfo.Format=";
char *META_2 = "\nEnd\n";

    if (sskfile && sskpath) {
	char *pubkey, *privkey;

        readKeys(sskfile, &pubkey, &privkey);
	key = malloc(strlen(privkey) + 20 + strlen(sskpath));
	strcpy(key, "freenet:SSK@");
	strcat(key, privkey);
	strcat(key, "/");
	while (*sskpath == '/') { sskpath++; }
	strcat(key, sskpath);

	free(pubkey);
	free(privkey);
    } else if (!key) {
	die("Must specify either key or sskfile and sskpath!");
    }

    if (meta && ct) {
	die("Metadata file and content type are mutually exclusive!");
    }
    if (ct) {
	metalen = strlen(META_1) + strlen(ct) + strlen(META_2);
	metadata = malloc(metalen + 1);
	strcpy(metadata, META_1);
	strcat(metadata, ct);
	strcat(metadata, META_2);
    }
    if (meta) {
	int fd;

	if (stat(meta, &sb)) {
	    die("Can't access metadata file!");
	}
	metalen = sb.st_size;
	metadata = malloc(metalen + 1);
	if (!metadata) {
	    die("malloc failed!");
	}

	if ((fd = open(meta, O_RDONLY)) < 0) {
	    die("Can't open metadata file!");
	}
	if (read(fd, metadata, metalen) < metalen) {
	    die("Can't read metadata file!");
	}
	close(fd);

	metadata[metalen] = 0;
    }

    if (!file) {
	die("file missing!");
    }

    if (fcpPutKeyFromFile(hfcp, key, file, metadata)) {
        fcpDestroyHandle(hfcp);
        die("Insert failed!");
    }

    if (sskfile && sskpath) {
	free(key);
    }
    if (metadata) {
	free(metadata);
    }
}

/** Parse the given string as a time interval:
 *   - either a plain number of seconds, or
 *   - a number of hours followed by the letter 'h', or
 *   - a number of days followed by the letter 'd'
 */
static unsigned int parseInterval(char *str) {
int i;
unsigned int mult = 1;

    if (!str) {
	die("parseInterval: Missing argument!");
    }
    for (i = 0; str[i]; i++) {
	if (str[i] == 'h') {
	    mult = 3600;
	    str[i--] = 0;
	} else if (str[i] == 'd') {
	    mult = 86400;
	    str[i--] = 0;
	}
    }

    return ((unsigned int) atoi(str)) * mult;
}

int main(int argc, char **argv) {
int i = 1, cmd = 0;
int dbr_offset = 0, dbr_increment = 0, dbr_period = 0, port = 8481, htl = 15;
char *target = NULL, *sskfile = NULL, *sskpath = NULL, *host = "localhost",
     *key = NULL, *ct = NULL, *meta = NULL;
HFCP *hfcp;

    me = argv[0];
    while (i < argc) {
	if (!strcmp(argv[i], "--genkeys")) {
	    if (cmd) { usage("Only one command allowed at a time!"); }
	    cmd = CMD_GENKEYS;
	} else if (!strcmp(argv[i], "--gendbr")) {
	    if (cmd) { usage("Only one command allowed at a time!"); }
	    if (i == argc - 1) { usage("Missing <target-URI>!"); }
	    cmd = CMD_GENDBR;
	    target = argv[++i];
	} else if (!strcmp(argv[i], "--redirect")) {
	    if (cmd) { usage("Only one command allowed at a time!"); }
	    if (i == argc - 1) { usage("Missing <target>!"); }
	    cmd = CMD_REDIRECT;
	    target = argv[++i];
	} else if (!strcmp(argv[i], "--insert")) {
	    if (cmd) { usage("Only one command allowed at a time!"); }
	    if (i == argc - 1) { usage("Missing <file>!"); }
	    cmd = CMD_INSERT;
	    target = argv[++i];
	} else if (!strcmp(argv[i], "--sskfile")) {
	    if (i == argc - 1) { usage("Missing <filename>!"); }
	    sskfile = argv[++i];
	} else if (!strcmp(argv[i], "--sskpath")) {
	    if (i == argc - 1) { usage("Missing <path>!"); }
	    sskpath = argv[++i];
	} else if (!strcmp(argv[i], "--dbr-offset")) {
	    if (i == argc - 1) { usage("Missing <time>!"); }
	    dbr_offset = parseInterval(argv[++i]);
	} else if (!strcmp(argv[i], "--dbr-increment")) {
	    if (i == argc - 1) { usage("Missing <time>!"); }
	    dbr_increment = parseInterval(argv[++i]);
	} else if (!strcmp(argv[i], "--dbr-period")) {
	    if (i == argc - 1) { usage("Missing <x>!"); }
	    dbr_period = atoi(argv[++i]);
	} else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
	    usage(NULL);
	} else if (!strcmp(argv[i], "--key")) {
	    if (i == argc - 1) { usage("Missing <key>!"); }
	    key = argv[++i];
	} else if (!strcmp(argv[i], "--content-type")) {
	    if (i == argc - 1) { usage("Missing <str>!"); }
	    ct = argv[++i];
	} else if (!strcmp(argv[i], "--metadata")) {
	    if (i == argc - 1) { usage("Missing <file>!"); }
	    meta = argv[++i];
	} else if (!strcmp(argv[i], "--version")) {
	    fprintf(stderr, "%s version %s\n", me, VERSION);
	    fprintf(stderr, "(C) 2003 Peter Conrad <conrad@unix-ag.uni-kl.de>\n");
	    fprintf(stderr, "This software may be used and / or distributed under the terms and conditions\n");
	    fprintf(stderr, "of the GNU General Public License version 2.\n");
	    exit(0);
	} else {
	    usage("Unknown option!");
	}
	i++;
    }

    if (getenv("FCP_HOST") != NULL) {
	host = getenv("FCP_HOST");
    }
    if (getenv("FCP_PORT") != NULL) {
	port = atoi(getenv("FCP_PORT"));
    }
    if (getenv("INSERT_HTL") != NULL) {
	htl = atoi(getenv("INSERT_HTL"));
    }

    if (fcpStartup(host, port, htl, 1, 1)) {
	die("fcpStartup failed!");
    }
    if (!(hfcp = fcpCreateHandle())) {
	die("fcpCreateHandle failed!");
    }
    fcpSetRmLocal(hfcp, 1);
    fcpSetSplitSize(4000000);

    if (cmd == CMD_GENKEYS) {
	genKey(hfcp, sskfile);
    } else if (cmd == CMD_REDIRECT) {
	genRedirect(hfcp, sskfile, sskpath, target);
    } else if (cmd == CMD_GENDBR) {
	genDBR(hfcp, sskfile, sskpath, target, dbr_increment, dbr_offset,
	       dbr_period);
    } else if (cmd == CMD_INSERT) {
	if (key == NULL && (sskfile == NULL || sskpath == NULL)
		|| key != NULL && (sskfile != NULL || sskpath != NULL)) {
	    die("--insert requires either --key or --sskfile and --sskpath!");
	}
	if (ct != NULL && meta != NULL) {
	    die("--content-type and --metadata are mutually exclusive!");
	}
	insertFile(hfcp, sskfile, sskpath, key, target, ct, meta);
    } else {
	usage("Missing command!");
    }

    fcpDestroyHandle(hfcp);

    exit(0);
}
 
/** Write log messages to stderr */
static void fn_debug(char *msg) {
    write(STDERR_FILENO, msg, strlen(msg));
    fsync(STDERR_FILENO);
}

/** Callback function used by ezFCPlib
 */
int fcpLogCallback(int level, char *buf) {
char *severity = "";
 
    if (level == FCP_LOG_CRITICAL) {
        severity = "CRITICAL";
    } else if (level == FCP_LOG_NORMAL) {
        severity = "NORMAL";
    } else if (level == FCP_LOG_VERBOSE) {
        severity = "VERBOSE";
    } else if (level == FCP_LOG_DEBUG) {
        severity = "DEBUG";
    }
 
//    if (level == FCP_LOG_CRITICAL || level == FCP_LOG_NORMAL) {
        fn_debug("ezFCPlib [");
        fn_debug(severity);
        fn_debug("]: ");
        fn_debug(buf);
        fn_debug("\n");
//    }
 
    return 0;
}
