/* freenet-handler.c: 
 *
 ****************************************************************
 * Copyright (C) 2003 Peter Conrad
 * Based on local-handler.c (C) 2002, 2003 Tom Lord
 * 
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include "hackerlab/char/str.h"
#include "hackerlab/os/errno.h"
#include "hackerlab/os/errno-to-string.h"
#include "hackerlab/os/sys/wait.h"
#include "hackerlab/os/signal.h"
#include "hackerlab/arrays/ar.h"
#include "hackerlab/fmt/cvt.h"
#include "hackerlab/vu/safe.h"
#include "hackerlab/vu-network/url-socket.h"


#include "ftp-utils/with-ftp/utils.h"
#include "ftp-utils/with-ftp/with-ftp.h"
#include "ftp-utils/with-ftp/freenet-handler.h"

#include "ezFCPlib/ezFCPlib.h"

#include <time.h>

/* "root" entry representing the top directory
 */
static fn_entry *root = NULL;

/* Remote "current working directory".
 * 
 * (The entire path provided to with-ftp.)
 */
static t_uchar *remote_cwd = NULL;

/* Various parameters for fcpget
 */
static t_uchar *fcp_host = "localhost";
static unsigned int fcp_port = 8481;
static unsigned int fcp_htl = 15;
static unsigned int fcp_htlinc = 5;
static unsigned int fcp_retries = 2;

ftp_handler_t
freenet_init_handler (t_uchar ***new_env, t_uchar *site)
{
t_uchar *tmpfile;
int isDBR = 0;

  /* Check presence of a few environment variables */
  fn_debug_init();
  if (getenv("FCP_HOST")) {
    fcp_host = getenv("FCP_HOST");
  }
  fcp_port = parse_env_number(getenv("FCP_PORT"), fcp_port);
  fcp_htl = parse_env_number(getenv("FCP_HTL"), fcp_htl);
  fcp_htlinc = parse_env_number(getenv("FCP_HTL_INCR"), fcp_htlinc);
  fcp_retries = parse_env_number(getenv("FCP_RETRIES"), fcp_retries);
  if (fcpStartup(fcp_host, fcp_port, fcp_htl, 1, 1)) {
    with_ftp_panic("fcpStartup failed!");
  }

  /* Check that *site is a valid freenet key, i. e. either
   * SSK@<key>[/<path>] or CHK@<key>
   */
  if (str_cmp_prefix("SSK@", site) && str_cmp_prefix("CHK@", site)) {
    with_ftp_panic("Unrecognized key type. Only SSK and CHK supported.");
  }

  tmpfile = str_alloc_cat_many (0, getenv("HOME"), "/.with-ftp/tempfile.XXXXXX",
				str_end);
  if (!mkstemp(tmpfile) || !tmpfile[0]) {
    fn_debug("mkstemp failed: ");
    fn_debug(strerror(errno));
    fn_debug("\n");
    with_ftp_panic("mkstemp failed!");
  }

  if (fn_download(site, tmpfile, 1, fcp_htl, fcp_htlinc, fcp_retries)) {
    unlink(tmpfile);
    with_ftp_panic("Failed to download metadata for key!");
  }

  /* Download the given freenet key's metadata, following Redirects and DBR */
  do {
    /* Parse the resulting file into a file & directory structure (represented
     * internally */
    if (root) { freeEntry(root); }
    root = parseMetadata(tmpfile);

    /* How to recognize "intermediate" keys that serve only to redirect to the
     * "real" content, like DBR keys?
     * We assume a key to be an intermediate key if it only has a default
     * document, which is a DateRedirect or Redirect.
     * While this is not perfect, it is good enough for us: if there are no
     * named documents in the metadata there's nothing to download for us,
     * anyway.
     */
    if (root->e.dir.default_entry
	    && (root->e.dir.default_entry->type == FN_TYPE_DBR
		|| root->e.dir.default_entry->type == FN_TYPE_REDIRECT)
	    && !root->e.dir.first_entry) {
      fn_entry *e = root->e.dir.default_entry;
      t_uchar *redir;
      t_uchar date[32];
      int offset = time(NULL);

      if (e->type == FN_TYPE_DBR) {
        if (!e->e.dbr.increment) { e->e.dbr.increment = 86400; }
        offset -= (offset - e->e.dbr.offset) % e->e.dbr.increment;
        cvt_long_to_hex(date, (long) offset);
        redir = str_alloc_cat_many(0, e->e.dbr.URI_base, date, "-",
				   e->e.dbr.URI_string, str_end);
      } else {
	redir = str_save(0, e->e.redirect.target);
      }
      if (fn_download(redir, tmpfile, 1, fcp_htl, fcp_htlinc, fcp_retries)) {
	unlink(tmpfile);
	with_ftp_panic("Failed to download metadata for redirected key!");
      }
      lim_free(0, redir);
      isDBR = 1;
    } else {
      isDBR = 0;
    }
  } while (isDBR);

  /* Remove temp file, initialize a few things */
  remote_cwd = str_save(0, "/");
  unlink(tmpfile);
  lim_free(0, tmpfile);

  return freenet_handle_command;
}

void
freenet_handle_command (int conn_in, int conn_out, t_uchar *cmd, int cmd_len)
{
  int errn;
  t_uchar * message;
  int cmd_status = 0;

  fn_debug("Received command: ");
  message = str_save_n(0, cmd, cmd_len);
  fn_debug(message);
  lim_free(0, message);

  cmd_len = trim_crnl (cmd, cmd_len);

  /* cmds not requiring a data port */

  if (!str_cmp_prefix ("cd ", cmd))
    {
      /* "cd <path>" */
      t_uchar *path, *cpath;

      /* check if it's relative to cwd or absolute */
      if (cmd[3] == '/') {
	path = str_save(0, cmd + 3);
      } else {
	path = str_alloc_cat_many(0, remote_cwd, "/", cmd + 3, str_end);
      }
      cpath = canonicalize_path(path);

      /* Check if it's known */
      if (findPath(root, cpath, 0)) {
        /* set new remote cwd */
        remote_cwd[0] = 0;
        remote_cwd = str_realloc_cat(0, remote_cwd, cpath);
        message = str_save (0, "chdir ok");
        cmd_status = 0;
      } else {
        message = str_save (0, "chdir failed: no such directory");
        cmd_status = -1;
      }
      lim_free(0, path);
    }
  else if (!str_cmp_prefix ("cdup", cmd))
    {
      /* "cdup" */
      if (!str_cmp("/", remote_cwd)) {
	message = str_save (0, "Already at root dir");
      } else {
        t_uchar *last = str_chr_rindex(remote_cwd, '/');
	if (!last) {
	  with_ftp_panic ("Assertion failed: remote_cwd doesn't contain any /es\n");
	}
	if (last == remote_cwd) { last++; }
	*last = 0;
	message = str_save (0, "chdir .. ok");
      }
      cmd_status = 0;
    }
  else if (!str_cmp_prefix ("delete ", cmd))
    {
      /* "delete <path>" */
      message = str_save (0, "Unsupported cmd 'delete' ignored.");
      cmd_status = 0;
    }
  else if (!str_cmp_prefix ("home", cmd))
    {
      /* "home" */
      str_cpy(remote_cwd, "/");
      message = str_save (0, "home ok");
      cmd_status = 0;
    }
  else if (!str_cmp_prefix ("rmdir ", cmd))
    {
      /* "rmdir <path>" */
      message = str_save (0, "Unsupported cmd 'rmdir' ignored.");
      cmd_status = 0;
    }
  else if (!str_cmp_prefix ("mkdir ", cmd))
    {
      /* "mkdir <path>" */
      message = str_save (0, "Unsupported cmd 'mkdir'.");
      cmd_status = -1;
    }
  else if (!str_cmp_prefix ("noop", cmd))
    {
      /* "noop" */
      cmd_status = 0;
      message = str_save (0, "noop ok");
    }
  else if (!str_cmp_prefix ("pwd", cmd))
    {
      if ((0 > printfmt (&errn, conn_out, "OK %lu\n%s\n", (unsigned long)(1 + str_length (remote_cwd)), remote_cwd))
	  || (0 > vfdbuf_flush (&errn, conn_out)))
	{
	  if (errn != EPIPE)
	    with_ftp_panic ("I/O error in local_handle_cmd (pwd OK)\n");
	}
      return;
    }
  else if (!str_cmp_prefix ("rnfr ", cmd))
    {
      message = str_save (0, "Unsupported cmd 'rnfr' ignored.");
      cmd_status = 0;
    }
  else if (!str_cmp_prefix ("rnto ", cmd))
    {
      message = str_save (0, "Unsupported cmd 'rnto'.");
      cmd_status = -1;
    }  


  /* cmds requiring a data port */

  else
    {
      t_uchar * host;
      t_uchar * comma;
      t_uchar * port;
      t_uchar * args;
      t_ulong h;
      t_uint p;
      int data_fd;

      /* <cmd> host,port args...
       * or
       * <cmd> pasv args...
       */

      host = str_chr_index (cmd, ' ');
      if (!host)
	goto bad_cmd;

      ++host;

      if (!str_cmp_prefix ("pasv ", host))
	{
	  /* passive mode */
	  /* block postconditions: args points at the list of args, data_fd ready */
	  if ((0 > printfmt (&errn, conn_out, "%s\n", my_data_url))
	      || (0 > vfdbuf_flush (&errn, conn_out)))
	    {
	      if (errn == EPIPE)
		return;
	      else
		with_ftp_panic ("I/O error in local_handle_cmd (PASV url send)\n");
	    }
	  data_fd = safe_open (my_data_url, O_RDWR, 0);
	  args = host + 5;	/* "pasv " */
	}
      else
	{
	  /* non-passive mode */
	  /* block postconditions: args points at the list of args, data_fd ready */

	  comma = str_chr_index (host, ',');
	  if (!comma)
	    goto bad_cmd;

	  if (cvt_decimal_to_ulong (&errn, &h, host, comma - host))
	    goto bad_cmd;

	  port = comma + 1;

	  args = str_chr_index (port, ' ');
	  if (!args)
	    goto bad_cmd;

	  if (cvt_decimal_to_uint (&errn, &p, port, args - port))
	    goto bad_cmd;

	  args += 1;

	  data_fd = url_inet_client_addr (&errn, h, p);
	  if (data_fd < 0)
	    {
	      message = str_save (0, errno_to_string (errn));
	      cmd_status = -1;
	      goto cmd_done;
	    }
	}

      if (!str_cmp_prefix ("get ", cmd))
	{
	  int file_fd;
	  t_uchar *tmpfile;
	  t_uchar *key;
	  fn_entry *dir;

	  if (args[0] == '/') {
	    t_uchar *slash = str_chr_rindex(args, '/');
	    if (slash == args) {
	      dir = root;
	    } else {
	      *slash = 0;
	      dir = findPath(root, args, 0);
	    }
	    args = slash+1;
	  } else {
	    dir = findPath(root, remote_cwd, 0);
	  }

	  dir = dir->e.dir.first_entry;
	  while (dir && strcmp(dir->name, args)) {
	    dir = dir->next;
	  }
	  if (!dir) {
	    message = str_save (0, "file not found");
	    cmd_status = -1;
	    goto cmd_done;
	  }
	  if (dir->type == FN_TYPE_DIR) {
	    if (dir->e.dir.default_entry) {
	      dir = dir->e.dir.default_entry;
	    } else {
	      message = str_save (0, "file is dir without default entry!");
	      cmd_status = -1;
	      goto cmd_done;
	    }
	  }
	  if (dir->type != FN_TYPE_REDIRECT) {
	    message = str_save (0, "unsupported part type!?");
	    cmd_status = -1;
	    goto cmd_done;
	  }
	  key = dir->e.redirect.target;
	  tmpfile = str_alloc_cat_many (0, getenv ("HOME"), "/.with-ftp/tempfile.XXXXXX", str_end);
	  if (!mkstemp(tmpfile) || !tmpfile[0]) {
	    lim_free(0, tmpfile);
	    fn_debug("mkstemp failed: ");
	    fn_debug(strerror(errno));
	    fn_debug("\n");
	    message = str_save (0, "mkstemp failed");
	    cmd_status = -1;
	    goto cmd_done;
	  }
	  if (fn_download(key, tmpfile, 0, fcp_htl, fcp_htlinc,fcp_retries)){
	    lim_free(0, tmpfile);
	    message = str_save (0, "download failed");
	    cmd_status = -1;
	    goto cmd_done;
	  }
	  fn_debug("Downloaded tmpfile\n");

	  file_fd = vu_open (&errn, tmpfile, O_RDONLY, 0);
	  if (file_fd < 0)
	    {
	      int ign;
	      unlink(tmpfile);
	      lim_free(0, tmpfile);

	    bad_data_cmd:
	      vu_close (&ign, data_fd);
	      message = str_save (0, errno_to_string (errn));
	      cmd_status = -1;
	      goto cmd_done;
	    }
	  else
	    {
	      message = str_save (0, "file opened for get");
	      fn_debug("File opened\n");
	      if ((0 > printfmt (&errn, conn_out, "OK %lu\n %s\n", (unsigned long)(2 + str_length (message)), message))
		  || (0 > vfdbuf_flush (&errn, conn_out)))
		{
		  if (errn == EPIPE) {
		    unlink(tmpfile);
		    lim_free(0, tmpfile);
		    goto bad_data_cmd;
		  } else
		    with_ftp_panic ("I/O error in local_handle_cmd (get OK)\n");
		}
	    }

	  {
	    char buf[4096];
	    size_t got;

	    while (1)
	      {
	        fn_debug("Reading file...\n");
		got = vu_read_retry (&errn, file_fd, buf, sizeof (buf));
		if (   (got < 0)
		    || (   (got > 0)
			&& (got != vu_write_retry (&errn, data_fd, buf, got))))
		  {
		    safe_close (file_fd);
		    unlink(tmpfile);
		    lim_free(0, tmpfile);
		    goto bad_data_cmd;
		  }
		else if (got == 0)
		  {
		    if (   (0 > vu_close (&errn, file_fd))
			|| (0 > vu_close (&errn, data_fd))) {
		      unlink(tmpfile);
		      lim_free(0, tmpfile);
		      goto bad_data_cmd;
		    }
		    break;
		  }
	      }
	  }

	  unlink(tmpfile);
	  lim_free(0, tmpfile);

	  message = str_save (0, "get cmd ok");
	  cmd_status = 0;
	}
      else if (!str_cmp_prefix ("put ", cmd))
	{
	  message = str_save (0, "put not supported");
	  cmd_status = 1;
	}
      else if (!str_cmp_prefix ("nlist ", cmd)
		|| !str_cmp_prefix ("list ", cmd))
	{
	  fn_entry *dir;

	  if (!str_length(args)) {
	    dir = findPath(root, remote_cwd, 0);
	  } else if (args[0] == '/') {
	    dir = findPath(root, args, 0);
	  } else {
	    t_uchar *path = str_alloc_cat_many(0, remote_cwd, "/", args,
					       str_end);
	    dir = findPath(root, path, 0);
	    lim_free(0, path);
	  }

	  if (!dir) {
	    int ign;
	    vu_close (&ign, data_fd);
	    message = str_save (0, "directory not found");
	    cmd_status = -1;
	    goto cmd_done;
	  }
	  message = str_save (0, "dir opened for nlist");
	  if ((0 > printfmt (&errn, conn_out, "OK %lu\n %s\n", (unsigned long)(2 + str_length (message)), message))
		  || (0 > vfdbuf_flush (&errn, conn_out)))
	    {
	      if (errn == EPIPE)
	        goto bad_data_cmd;
	      else
	        with_ftp_panic ("I/O error in local_handle_cmd (nlist OK)\n");
	    }

	  dir = dir->e.dir.first_entry;
	  while (dir) {
	    if (0 > printfmt (&errn, data_fd, "%s\r\n", dir->name))
	      {
	        goto bad_data_cmd;
	      }
	    dir = dir->next;
	  }

	  if (0 > vu_close (&errn, data_fd))
	    goto bad_data_cmd;
	  message = str_save (0, "nlist cmd ok");
	  cmd_status = 0;
	}
      else
	{
	bad_cmd:
	  safe_printfmt (2, "with-ftp: bad command recieved (%s)\n", cmd);
	  with_ftp_panic ("with-ftp protocol error");
	}
    }
      
 cmd_done:
  {
    int ign;
    printfmt (&ign, conn_out, "%s %lu\n %s\n", ((cmd_status < 0) ? "ERROR" : "OK"), (unsigned long)(2 + str_length (message)), message);
    vfdbuf_flush (&ign, conn_out);
    lim_free (0, message);
  }
}

/* tag: freenet-handler.c 9ca625b11c75042bb6e00413376419d2
 */
