/* freenet-utils.c: 
 *
 ****************************************************************
 * Copyright (C) 2003 Peter Conrad
 * 
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include "hackerlab/char/str.h"
#include "hackerlab/os/errno.h"
#include "hackerlab/os/errno-to-string.h"
#include "hackerlab/os/sys/wait.h"
#include "hackerlab/fmt/cvt.h"
#include "hackerlab/vu/safe.h"

#include "ftp-utils/with-ftp/with-ftp.h"

#include "ezFCPlib/ezFCPlib.h"

#include "freenet-utils.h"

#include <string.h>

#ifdef DEBUG_FN
static int debug_fd = -1;

void fn_debug(t_uchar *msg) {
    if (debug_fd >= 0) {
	safe_write(debug_fd, msg, str_length(msg));
	safe_fsync(debug_fd);
    }
}

void fn_debug_init() {
    t_uchar *env = getenv("FN_DEBUG");
    if (env) {
        debug_fd = safe_open(env, O_RDWR | O_CREAT | O_APPEND, 0666);
    }
}
#endif

/** Callback function used by ezFCPlib
 */
int fcpLogCallback(int level, char *buf) {
char *severity = "";

    if (level == FCP_LOG_CRITICAL) {
	severity = "CRITICAL";
    } else if (level == FCP_LOG_NORMAL) {
	severity = "NORMAL";
    } else if (level == FCP_LOG_VERBOSE) {
	severity = "VERBOSE";
    } else if (level == FCP_LOG_DEBUG) {
	severity = "DEBUG";
    }

    fn_debug("ezFCPlib [");
    fn_debug(severity);
    fn_debug("]: ");
    fn_debug(buf);
    fn_debug("\n");

    return 0;
}

/** If str is NULL return default, otherwise parse str into an unsigned int
 *  and return that. If the parsing fails, panic.
 */
unsigned int parse_env_number(t_uchar *str, unsigned int def_value) {
int rc, errn;
unsigned int value;

    if (!str) { return def_value; }
    rc = cvt_decimal_to_uint(&errn, &value, str, str_length(str));
    if (!rc) { return value; }

    with_ftp_panic("Failed to parse environment value!");
    return 0; /* only for suppressing a compiler warning */
}

/* Download (get) a given key from freenet into a given file. If meta
 * is true, only the metadata is downloaded, otherwise the key's content.
 * If usewait is true, the spawned subprocess will be waited for using 
 * waitpid, otherwise we rely one some other routine initializing a signal
 * handler that siglongjmp's to sigchld_jump_tool.
 * Returns 0 on success, otherwise an error code.
 */
int fn_download(t_uchar *key, t_uchar *target, int meta, int htl, int htlinc,
		int retries) {
int j = 0, success = 0;
HFCP *handle;
char *metadata;

    fn_debug("Downloading ");
    fn_debug(key);
    fn_debug(meta ? " (metadata) into " : " (content) into ");
    fn_debug(target);
    fn_debug("\n");

    handle = fcpCreateHandle();
    if (!handle) {
        fn_debug("fcpCreateHandle failed!?");
	return 1;
    }
    do {
	fcpSetHtl(handle, htl);
	fcpRawMode(handle, meta ? 1 : 0);
	/* Great. ezFCPlib documentation is lying wrt metadata. */
	if (!fcpGetKeyToFile(handle, key, target, &metadata)) {
	    /* Success */
	    success = 1;
	    if (meta) {
		if (handle->rawMetadata) {
		    int metaLen=handle->conn.response.body.datafound.metaLength;
		    /* write metadata to target file */
	            int meta_fd = safe_open(target, O_WRONLY, 0600);
	            if (meta_fd < 0) {
		        fn_debug("Can't write metadata file!\n");
		        with_ftp_panic("Can't write metadata file!");
	            }
		    if (safe_write(meta_fd, handle->rawMetadata, metaLen)
			    != metaLen) {
		        fn_debug("Short write!\n");
		        with_ftp_panic("Short write!");
		    }
		    close(meta_fd);
		} else {
		    fn_debug("No metadata?!\n");
		    success = -1;
		}
	    }
	} else {
	    /* Failure - retry */
	    htl += htlinc;
	}
    } while (j++ < retries && !success);

    fcpDestroyHandle(handle);

    fn_debug(success > 0 ? "Download successful\n" : "Download failed\n");

    return success > 0 ? 0 : -1;
}

/* Allocate a new fn_entry structure and initialize most fields to 0
 */
static fn_entry *createEntry() {
fn_entry *e;

    e = lim_malloc(0, sizeof(fn_entry));
    if (!e) {
	fn_debug("Can't alloc memory!\n");
	with_ftp_panic("Can't alloc memory!");
    }
    e->name = NULL;
    e->next = NULL;
    e->type = 0;
    // clearing dir also clears the other union members
    e->e.dir.fullpath = NULL;
    e->e.dir.parent = NULL;
    e->e.dir.first_entry = NULL;
    e->e.dir.default_entry = NULL;

    return e;
}

/* Free an fn_entry structure, if necessary recursing through subdirs.
 * Return e->next
 */
fn_entry *freeEntry(fn_entry *e) {
fn_entry *n;

    if (e->name) { lim_free(0, e->name); }
    if (e->type == FN_TYPE_DIR) {
	if (e->e.dir.fullpath) { lim_free(0, e->e.dir.fullpath); }
	n = e->e.dir.first_entry;
	while (n) { n = freeEntry(n); }
	if (e->e.dir.default_entry) { freeEntry(e->e.dir.default_entry); }
    } else if (e->type == FN_TYPE_DBR) {
	if (e->e.dbr.URI_base) { lim_free(0, e->e.dbr.URI_base); }
	if (e->e.dbr.URI_string) { lim_free(0, e->e.dbr.URI_string); }
    } else if (e->type == FN_TYPE_REDIRECT) {
	if (e->e.redirect.target) { lim_free(0, e->e.redirect.target); }
    }
    n = e->next;
    lim_free(0, e);
    return n;
}

/* Read a line (terminated by '\n') from the given filedescriptor into the
 * given buffer, terminating it with '\0'. At most n-1 characters are read.
 * The line terminator, if present, is removed. If at least one character
 * (which could be '\n') has been read into buffer, 1 is returned. Otherwise,
 * on end of file 0 is returned, or -1 on error.
 */
#ifndef DEBUG_FN
static
#endif
int readLine(t_uchar *buffer, int fd, int n) {
int i = 0, rc;

    do {
	buffer[i] = 0;
	rc = safe_read(fd, buffer + i, 1);
	if (rc == 0) {
	    return (i > 0) ? 1 : 0;
	}
	if (rc < 0) {
	    return -1;
	}
	if (buffer[i] == '\n') {
	    buffer[i] = 0;
	    return 1;
	}
    } while (i++ < n-2);
    buffer[i] = 0;
    return 1;
}

/* Check if the name of **e matches. If **e is not a dir entry, create a
 * new one, make the old **e its default entry, then set **e to the new
 * entry.
 * Return *e if the name matches, otherwise NULL
 */
#ifndef DEBUG_FN
static
#endif
fn_entry *checkPathElement(fn_entry **e, t_uchar *name) {
    if (!strcmp((*e)->name, name)) {
	if ((*e)->type != FN_TYPE_DIR) {
	    fn_entry *tmp = createEntry();
	    tmp->next = (*e)->next;
	    tmp->e.dir.default_entry = *e;
	    *e = tmp;
	    tmp->type = FN_TYPE_DIR;
	    tmp->name = str_save(0, name);
	}
	return *e;
    }

    return NULL;
}

/* Follow the given path from the given root entry to the entry representing
 * the entity specified by path.
 * If create is true assume that path is specifying a "directory" and create
 * any missing dir entries, as in "mkdir -p".
 */
fn_entry *findPath(fn_entry *root, t_uchar *path, int create) {
t_uchar *p, *p1, *slash;
fn_entry *cur = root, *tmp, *tmp2;

    if (!str_cmp(path, "/")) { return root; }

    p = p1 = str_save(0, path);
    if (*p1 == '/') { p1++; }

    do {
	slash = str_chr_index(p1, '/');
	if (slash) { *slash = 0; }
	if (!cur->e.dir.first_entry) {
	    if (!create) { return NULL; }
	    cur->e.dir.first_entry = createEntry();
	    cur->e.dir.first_entry->e.dir.parent = cur;
	    cur = cur->e.dir.first_entry;
	    cur->type = FN_TYPE_DIR;
	    cur->name = str_save(0, p1);
	} else {
	    tmp = checkPathElement(&(cur->e.dir.first_entry), p1);
	    if (tmp) {
		tmp->e.dir.parent = cur;
		cur = tmp;
	    } else {
		tmp2 = cur->e.dir.first_entry;
		while (tmp2->next 
			&& !(tmp = checkPathElement(&(tmp2->next), p1))) {
		    tmp2 = tmp2->next;
		}
		if (tmp) {
		    cur = tmp;
		} else {
		    if (!create) { return NULL; }
		    tmp2->next = createEntry();
		    tmp2->next->e.dir.parent = cur;
		    cur = tmp2->next;
		    cur->type = FN_TYPE_DIR;
		    cur->name = str_save(0, p1);
		}
	    }
	}
	p1 = slash + 1;
    } while (slash);

    lim_free(0, p);

    return cur;
}

/* Parse the given metadata file into a new fn_entry structure.
 */
fn_entry *parseMetadata(t_uchar *meta) {
int	 meta_fd, state = 0, rc = 0, errn;
char	 readbuf[512];
fn_entry *next = NULL, *targetDir = NULL, *root;

    meta_fd = safe_open(meta, O_RDONLY, 0444);
    if (meta_fd < 0) {
	fn_debug("Can't open metadata file!\n");
	with_ftp_panic("Can't open metadata file!");
    }
    root = createEntry();
    root->type = FN_TYPE_DIR;
    root->e.dir.fullpath = str_save(0, "/");

    while (state >= 0 && (rc=readLine(readbuf, meta_fd, sizeof(readbuf))) > 0) {
	fn_debug(readbuf);
	if (state == 1 
		&& (!str_cmp(readbuf, "EndPart") || !str_cmp(readbuf, "End"))) {
	    state = !str_cmp(readbuf, "EndPart") ? 3 : -1;
	    if (!next->name) {
		// this is the default document of the root dir
		if (root->e.dir.default_entry) {
		    freeEntry(root->e.dir.default_entry);
		}
		root->e.dir.default_entry = next;
	    } else {
		// append next to the targetDir's children
		if (!targetDir->e.dir.first_entry) {
		    targetDir->e.dir.first_entry = next;
		} else {
		    targetDir = targetDir->e.dir.first_entry;
		    while (targetDir->next
			   && str_cmp(targetDir->name, next->name)) {
			targetDir = targetDir->next;
		    }
		    if (!str_cmp(targetDir->name, next->name)) {
			fn_debug(" - ignoring duplicate");
			freeEntry(next);
		    } else {
			targetDir->next = next;
		    }
		}
	    }
	    next = NULL;
	} else if (!str_cmp(readbuf, "Document")) {
	    next = createEntry();
	    targetDir = root;
	    state = 1;
	} else if (!str_cmp_prefix("DateRedirect.", readbuf) && state == 1
		   && (next->type == 0 || next->type == FN_TYPE_DBR)) {
	    next->type = FN_TYPE_DBR;
	    if (!str_cmp_prefix("DateRedirect.Target=", readbuf)) {
		int start = 20;
		if (!str_cmp_prefix("freenet:", readbuf + start)) {
		    start += 8;
		}
		if (next->e.dbr.URI_base) {
		    lim_free(0, next->e.dbr.URI_base);
		    next->e.dbr.URI_base = NULL;
		}
		if (next->e.dbr.URI_string) {
		    lim_free(0, next->e.dbr.URI_string);
		    next->e.dbr.URI_string = NULL;
		}
		if (!str_cmp_prefix("KSK@", readbuf + start)) {
		    next->e.dbr.URI_base = str_save(0, "KSK@");
		    next->e.dbr.URI_string = str_save(0, readbuf + start + 4);
		} else if (!str_cmp_prefix("SSK@", readbuf + start)) {
		    t_uchar *slash = str_chr_index(readbuf + start, '/');
		    if (slash) {
			int len = 0;
			t_uchar *t = readbuf + start;
			slash++;
			while (t++ != slash) { len++; }
		        next->e.dbr.URI_base = str_save_n(0, readbuf + start,
							  len);
		        next->e.dbr.URI_string = str_save(0, slash);
		    }
		}
	    } else if (!str_cmp_prefix("DateRedirect.Increment=", readbuf)) {
		cvt_hex_to_uint(&errn, &next->e.dbr.increment, readbuf + 23,
				str_length(readbuf + 23));
	    } else if (!str_cmp_prefix("DateRedirect.Offset=", readbuf)) {
		cvt_hex_to_uint(&errn, &next->e.dbr.offset, readbuf + 20,
				str_length(readbuf + 20));
	    } else {
		fn_debug(" - ignoring unknown DateRedirect key");
	    }
	} else if (!str_cmp_prefix("Redirect.Target=", readbuf) && state == 1
		   && (next->type == 0 || next->type == FN_TYPE_REDIRECT)) {
	    next->type = FN_TYPE_REDIRECT;
	    if (next->e.redirect.target) {
		lim_free(0, next->e.redirect.target);
	    }
	    next->e.redirect.target = str_save(0, readbuf + 16);
	} else if (!str_cmp_prefix("Name=", readbuf) && state == 1) {
	    t_uchar *slash = str_chr_rindex(readbuf + 5, '/');
	    if (next->name) {
		fn_debug(" - using duplicate name key");
		lim_free(0, next->name);
	    }
	    if (!slash) {
		next->name = str_save(0, readbuf + 5);
	        targetDir = root;
	    } else {
		next->name = str_save(0, slash + 1);
		*slash = 0;
	        targetDir = findPath(root, readbuf + 5, 1);
	    }
	} else {
	    fn_debug(" - ignored");
	}
	fn_debug("\n");
    }

    if (rc < 0) {
	fn_debug("Error reading meta data: ");
	fn_debug(errno_to_string(errno));
	fn_debug("\n");
	with_ftp_panic("Error reading meta data!");
    }

    return root;
}

/* Canonicalize a path name:
 *  - replace all // with a single /
 *  - replace all /./ with a single /
 *  - remove trailing /.
 *  - replace all <dir>/../ with /
 *  - replace all /../ (at the beginning of the pathname) with /
 *  - replace all <path>/<dir>/.. (at the end of the pathname) with <path>
 * Return a pointer into the given argument string. The argument string may
 * be modified.
 */
t_uchar *canonicalize_path(t_uchar *path) {
  t_uchar *path2;
  t_uchar *path3;

  fn_debug("Canonicalize_path: '");
  fn_debug(path);
  fn_debug("'\n");

  if (*path != '/') {
    with_ftp_panic("Assertion failed: path must start with a '/'!");
  }

  /* Remove double slashes */
  path2 = path;
  path3 = path + 1;
  while (*path3) {
    if (*path2 != '/' || *path3 != '/') {
      *++path2 = *path3++;
    } else {
      path3++;
    }
  }
  *++path2 = 0;
  /* At this point, path2 points to the end of the string (NUL). */
  /* Remove trailing /, if present */
  if (*--path2 == '/') {
    *path2 = 0;
  } else {
    path2++;
  }

  /* Remove trailing /.'s */
  while (path2 > path && *--path2 == '.' && *--path2 == '/') {
    if (path2 == path) {
      *++path2 = 0;
    } else {
      *path2 = 0;
    }
  }

  /* Replace all /./ with / */
  path2 = path;
  path3 = path;
  while (*path3) {
    if (*path3 == '/' && path3[1] == '.' && path3[2] == '/') {
      path3 += 2;
    } else {
      *path2++ = *path3++;
    }
  }
  *path2 = 0;

  /* Handle /../ and /.. */
  do {
    /* First, handle /.. at the beginning of path */
    while (!str_cmp_prefix("/../", path)) { path += 3; }
    if (!str_cmp("/..", path)) { path[1] = 0; }

    /* Then, handle other occurrences of /.. */
    path2 = path;
    while (path2 && (str_cmp_prefix("/..", path2) ||
		     (path2[3] != '/' && path2[3]))) {
      path2++; 
      path2 = strstr(path2, "/..");
    }
    if (path2) {
      /* Here, path2 points to either the first /../ or /.. at the end of path.
       * Go back to the previous / and copy everything after the /.. to there */
      path3 = path2;
      while (*--path3 != '/') ;
      path2 += 3;
      while ((*path3++ = *path2++));
    }
  } while (path2);

  fn_debug("Canonical result: '");
  fn_debug(path);
  fn_debug("'\n");

  return path;
}

/* tag: freenet-utils.c c4dfdfee74896f5bb2d3cda18e115a6a
 */
